#include <stdio.h>
#include <jni.h>
#include "java_vrpn.h"
#include "vrpn_Button.h"
#include "vrpn_ButtonRemote.h"

JavaVM* jvm = NULL;
jclass jclass_vrpn_ButtonRemote = NULL;
jfieldID jfid_vrpn_ButtonRemote_native_button = NULL;

// This is called when the Java Virtual Machine loads this library
//  and sets some global references that are used elsewhere.
JNIEXPORT jint JNICALL JNI_OnLoad( JavaVM* jvm, void* reserved )
{
  //////////////
  // set the jvm reference
  ::jvm = jvm;

  ///////////////
  // get the JNIEnv pointer
  JNIEnv* env;
  if( jvm->GetEnv( (void**) &env, JAVA_VRPN_JNI_VERSION ) != JNI_OK )
  {
    printf( "Error loading vrpn ButtonRemote native library.\n" );
    return JNI_ERR;
  }
  
  ///////////////
  // get the jclass reference to vrpn.ButtonRemote
  jclass cls = env->FindClass( "vrpn/ButtonRemote" );
  if( cls == NULL )
  {
    printf( "Error loading vrpn ButtonRemote native library "
            "while trying to find class vrpn.ButtonRemote.\n" );
    return JNI_ERR;
  }

  // make a global reference so the class can be referenced later.
  // this must be a _WEAK_ global reference, so the class will be
  // garbage collected, and so the JNI_OnUnLoad handler will be called
  jclass_vrpn_ButtonRemote = (jclass) env->NewWeakGlobalRef( cls );
  if( jclass_vrpn_ButtonRemote == NULL )
  {
    printf( "Error loading vrpn ButtonRemote native library "
            "while setting up class vrpn.ButtonRemote.\n" );
    return JNI_ERR;
  }

  ////////////////
  // get a jfid field id reference to the "native_button" 
  // field of class vrpn.ButtonRemote.
  // field ids do not have to be made into global references.
  jfid_vrpn_ButtonRemote_native_button 
    = env->GetFieldID( jclass_vrpn_ButtonRemote, "native_button", "I" );
  if( jfid_vrpn_ButtonRemote_native_button == NULL )
  {
    printf( "Error loading vrpn ButtonRemote native library "
            "while looking into class vrpn.ButtonRemote.\n" );
    return JNI_ERR;
  }
  
  return JAVA_VRPN_JNI_VERSION;
} // end JNI_OnLoad


JNIEXPORT void JNICALL JNI_OnUnload( JavaVM* jvm, void* reserved )
{
  // get the JNIEnv pointer
  JNIEnv* env;
  if( jvm->GetEnv( (void**) &env, JAVA_VRPN_JNI_VERSION ) != JNI_OK )
  {
    printf( "Error unloading vrpn ButtonRemote native library.\n" );
    return;
  }

  // delete the global reference to the vrpn.ButtonRemote class
  env->DeleteWeakGlobalRef( jclass_vrpn_ButtonRemote );
}


// This is the callback for vprn to notify us of a new button message
void handle_button_change( void* userdata, const vrpn_BUTTONCB info )
{
  if( jvm == NULL )
    return;

  /*  
  printf( "button change (C):  time:  %d.%d;  button:  %d;\n"
          "\tstate: %d;\n", info.msg_time.tv_sec, info.msg_time.tv_usec,
          info.button, info.state );
  */

  JNIEnv* env;
  jvm->AttachCurrentThread( (void**) &env, NULL );
  
  jobject jobj = (jobject) userdata;
  jclass jcls = env->GetObjectClass( jobj );
  jmethodID jmid = env->GetMethodID( jcls, "handleButtonChange", "(JJII)V" );
  if( jmid == NULL )
  {
    printf( "Warning:  vrpn_ButtonRemote library was unable to find the "
            "Java method \'handleButtonChange\'.  This may indicate a version mismatch.\n" );
    return;
  }
  env->CallVoidMethod( jobj, jmid, (jlong) info.msg_time.tv_sec, (jlong) info.msg_time.tv_usec,
                       (jint) info.button, (jint) info.state );

}


JNIEXPORT void JNICALL 
Java_vrpn_ButtonRemote_mainloop( JNIEnv* env, jobject jobj )
{
  // look up the button pointer
  jclass jcls = env->GetObjectClass( jobj );
  jfieldID jfid = env->GetFieldID( jcls, "native_button", "I" );
  if( jfid == NULL )
  {
    printf( "Error in native method \"mainloop\":  unable to ID native button field.\n" );
    return;
  }
  vrpn_Button_Remote* t = (vrpn_Button_Remote*) env->GetIntField( jobj, jfid );
  if( t <= 0 )  // this button is uninitialized or has been shut down already
    return;

  // now call mainloop
  t->mainloop( );
}


JNIEXPORT jboolean JNICALL 
Java_vrpn_ButtonRemote_init( JNIEnv* env, jobject jobj, jstring jname )
{
  printf( "in Java_ButtonRemote_init(...)\n" );

  // look up where to store the button pointer
  jclass jcls = env->GetObjectClass( jobj );
  jfieldID jfid = env->GetFieldID( jcls, "native_button", "I" );
  if( jfid == NULL )
  {
    printf( "Error in native method \"init\":  unable to ID native button field.\n" );
    return false;
  }

  // make a global reference to the Java ButtonRemote
  // object, so that it can be referenced in the function
  // handle_button_change(...)
  jobj = env->NewGlobalRef( jobj );

  // create the button
  const char* name = env->GetStringUTFChars( jname, NULL );
  vrpn_Button_Remote* t = new vrpn_Button_Remote( name );
  t->register_change_handler( jobj, handle_button_change );
  env->ReleaseStringUTFChars( jname, name );
  
  // now stash 't' in the jobj's 'native_button' field
  jint jt = (jint) t;
  env->SetIntField( jobj, jfid, jt );
  
  return true;
}



// this is only supposed to be called when the ButtonRemote
// instance is finalized for garbage collection
JNIEXPORT void JNICALL 
Java_vrpn_ButtonRemote_shutdownButton( JNIEnv* env, jobject jobj )
{

  // look up where to store the button pointer
  jclass jcls = env->GetObjectClass( jobj );
  jfieldID jfid = env->GetFieldID( jcls, "native_button", "I" );
  if( jfid == NULL )
  {
    printf( "Error in native method \"shutdownButton\":  unable to ID native button field.\n" );
    return;
  }

  // get the button pointer
  vrpn_Button_Remote* t = (vrpn_Button_Remote*) env->GetIntField( jobj, jfid );
  
  // unregister a handler and destroy the button
  if( t > 0 )
  {
    t->unregister_change_handler( jobj, handle_button_change );
    delete t;
  }

  // set the button pointer to -1
  env->SetIntField( jobj, jfid, -1 );

  // delete global reference to object (that was created in init)
  env->DeleteGlobalRef( jobj /* is this the right object? */ );


}




